/*
 * Memory area support.
 *
 * Copyright (C) 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <map>
#include <set>
#include <stack>

#include <l4/sys/types.h>

#include <systypes/base.h>



/* A memory area abstraction. */

class MemoryArea
{
protected:
  address_t _start, _end;

public:
  explicit MemoryArea()
  : _start(0), _end(0)
  {
  }

  explicit MemoryArea(address_t start, address_t end)
  : _start(start), _end(end)
  {
  }

  virtual ~MemoryArea()
  {
  }

  virtual MemoryArea *copy();

  /* Access to area properties. */

  virtual address_t area_start()
  { return _start; }

  virtual address_t area_end()
  { return _end; }

  virtual l4_umword_t flags()
  { return 0; }

  virtual l4_cap_idx_t dataspace()
  { return L4_INVALID_CAP; }

  virtual address_t dataspace_start()
  { return 0; }

  /* Return whether the area supports the given address. */

  virtual bool supports(address_t addr)
  { return (_start <= addr) && (addr < _end); }

  /* Return whether an area is reserved and therefore cannot be mapped. */

  virtual bool is_reserved()
  { return false; }

  /* Return whether the area provides a mapped region. */

  virtual bool is_mapped()
  { return false; }

  /* Support for populating areas. */

  virtual long add(MemoryArea &area);

  virtual long remove(MemoryArea &area);

  /* Support for finding regions. */

  virtual long find(address_t addr, MemoryArea **area, MemoryArea **parent);

  /* Support for finding areas. */

  virtual long find(address_t *start, address_t *size, map_flags_t flags,
                    unsigned char align, MemoryArea **area);
};



/* A reserved area abstraction. */

class ReservedMemoryArea : public MemoryArea
{
public:
  explicit ReservedMemoryArea(address_t start, address_t end)
  : MemoryArea(start, end)
  {
  }

  virtual MemoryArea *copy();

  /* Return whether an area is reserved and therefore cannot be mapped. */

  virtual bool is_reserved()
  { return true; }

  /* Support for finding areas. */

  virtual long find(address_t addr, MemoryArea **area, MemoryArea **parent);
};



/* Collection types. */

typedef std::map<address_t, MemoryArea *> MemoryAreaMap;
typedef std::set<MemoryArea *> MemoryAreas;



/* A memory area containing other areas. */

class AvailableMemoryArea : public MemoryArea
{
protected:
  MemoryAreaMap _areas;
  MemoryAreas _allocated;

public:
  explicit AvailableMemoryArea(address_t start, address_t end)
  : MemoryArea(start, end)
  {
  }

  virtual ~AvailableMemoryArea();

  virtual MemoryArea *copy();

  /* Support for populating areas. */

  virtual long add(MemoryArea &area);

  virtual long remove(MemoryArea &area);

  /* Support for finding areas. */

  virtual long find(address_t addr, MemoryArea **area, MemoryArea **parent);

  virtual long find(address_t *start, address_t *size, map_flags_t flags,
                    unsigned char align, MemoryArea **area);

  /* A recursive iterator over a memory area. */

  class iterator
  {
  protected:
    std::stack<MemoryAreaMap::iterator> _iterators, _ends;

    void ascend();

    void descend();

    void descend_all();

  public:
    explicit iterator();

    explicit iterator(MemoryAreaMap::iterator it,
                      MemoryAreaMap::iterator end);

    MemoryArea *operator *();

    iterator &operator ++();

    iterator &operator ++(int);

    bool operator ==(iterator other);

    bool operator !=(iterator other);

    /* Access to the underlying iterators. */

    MemoryAreaMap::iterator &area_iterator();

    MemoryAreaMap::iterator &area_end();
  };

  /* Iteration methods. */

  virtual iterator begin();

  virtual iterator end();

  virtual MemoryAreaMap::iterator areas_begin();

  virtual MemoryAreaMap::iterator areas_end();
};

/* vim: tabstop=2 expandtab shiftwidth=2
*/
